/*
** PAL Video using T3 and Output Compare interrupts
**
** Graphic page 
**
** Modified to produce properly sized and spaced pre-and post equalization pulses as well as vertical sync pulses
*/

#include <p24fj128ga010.h>
#include "Graphic.h"

// I/O definitions
#define SYNC    _LATG0  // output 
#define SDO     _RF8    // SPI1 SDO

// timing definitions for PAL video vertical state machine--------------------------------------------------------------------------------------------------
#define V_PAL   625    	// total number of lines composing a frame

#define VRES_1	VRES		// Total number of lines FIRST FIELD, active lines only
#define VRES_2	VRES		// Total number of lines SECOND FIELD, active lines only

#define VSYNC_N1  2     // V sync lines FIRST FIELD
#define VSYNC_N2  2     // V sync lines SECOND FIELD

#define VBLANK_N1  18  	// Number of blank lines FIRST FIELD
#define VBLANK_N2  18  	// Number of blank lines SECOND FIELD

#define PREEQ_N1   2    // pre equalization FIRST FIELD, 
#define PREEQ_N2   3    // pre equalization SECOND FIELD, odd
#define POSTEQ_N1  2    // post equalization FIRST FIELD, 
#define POSTEQ_N2  2    // post equalization SECOND FIELD, 

#define HALF_LINE_N1	1	// valid value is ONE, line 3 of 625
#define HALF_LINE_N2	1	// valid value is ONE, line 313 of 625

// definition of the sync state machine---------------------------------------------------------------------------------------------------------------------
#define SV_SYNC_1     	1
#define HALF_LINE_1		2
#define SV_POSTEQ_1   	3
#define SV_BLANKLINE_1	4	
#define SV_LINE_1     	5	
#define SV_PREEQ_1   	6
#define HALF_LINE_2		7
#define SV_SYNC_2     	8
#define SV_POSTEQ_2   	9
#define SV_BLANKLINE_2	10
#define SV_LINE_2     	11
#define SV_PREEQ_2   	0

// timing definitions for PAL video horizontal state machine------------------------------------------------------------------------------------------------
#define H_PAL  		1024    // total number of Tcy in a line (64us)
#define HSYNC_T  	75     	// Tcy in a horizontal synch pulse used during each scan line and vertical sync
#define BPORCH_T 	90     	// Tcy in a back porch 
#define SHORT_SYNC	38		// Tcy,duration of short sync pulses used during pre-and post equalization
#define PIX_T    	6      	// Tcy in each pixel, 2,3,4,5 or 6
//----------------------------------------------------------------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------------------------------------------------------------

#define _FAR __attribute__(( far))

int _FAR VMap[VRES * (HRES/16)];

volatile int *VPtr;
volatile int HCount, VCount, VState, HState;

// next state table
int VS[12] = { SV_SYNC_1,	HALF_LINE_1,	SV_POSTEQ_1,	SV_BLANKLINE_1,		SV_LINE_1,	SV_PREEQ_1,	HALF_LINE_2,	SV_SYNC_2,	SV_POSTEQ_2,	SV_BLANKLINE_2,	SV_LINE_2,	SV_PREEQ_2 	};
// next counter table
int VC[12] = { VSYNC_N1,	HALF_LINE_N1 ,	POSTEQ_N1, 		VBLANK_N1, 			VRES_1, 	PREEQ_N1,	HALF_LINE_N2,	VSYNC_N2,	POSTEQ_N2,		VBLANK_N2,		VRES_2,		PREEQ_N2	};


void _ISRFAST _T3Interrupt( void)
{
    // Start a Synch pulse
    SYNC = 0;

    // decrement the vertical counter
    VCount--;
    
    // vertical state machine
    switch ( VState) {

		case SV_SYNC_1://FIELD 1 - Lines 1,2
            // vertical sync pulse - OK
            OC3R = (H_PAL/2) - HSYNC_T;
            OC3CON = 0x0009;    // single event 

			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = H_PAL - HSYNC_T;		//
            OC2CON = 0x0009;    		// single event 
            break;

 		case HALF_LINE_1://FIELD 1 - Line 3
           // HALF LINE - OK
            OC3R = (H_PAL/2)- HSYNC_T;
            OC3CON = 0x0009;    // single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = (H_PAL/2) + SHORT_SYNC;	//
            OC2CON = 0x0009;    			// single event 
            break;

 		case SV_POSTEQ_1://FIELD 1 - Lines 4,5
            // Post equalization pulse - OK
            OC3R = SHORT_SYNC;
            OC3CON = 0x0009;    //single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = (H_PAL/2) + SHORT_SYNC;	// Relative to Timer 3
            OC2CON = 0x0009;    			// single event 
         	break;

		case SV_BLANKLINE_1://FIELD 1 - Lines vary
           // blank line - OK
            OC3R = HSYNC_T;
            OC3CON = 0x0009;    // single event 
            break;

	 	case SV_LINE_1://FIELD 1 - Lines vary
            // horizontal sync pulse - OK
            OC3R = HSYNC_T;
            OC3CON = 0x0009;    // single event 

            // activate OC4 for the SPI loading
            OC4R = HSYNC_T + BPORCH_T;
            OC4CON = 0x0009;    // single event 
			
            HCount = HRES/128;  // loads 8x16 bits at a time 
            break; 
     
        case SV_PREEQ_1://FIELD 1 - Lines 311,312
           	// Pre equalization pulse - OK
            OC3R = SHORT_SYNC;
            OC3CON = 0x0009;    //single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = (H_PAL/2) + SHORT_SYNC;	//Relative to Timer 3
            OC2CON = 0x0009;    			// single event 

         	break;
            
        case HALF_LINE_2://FIELD 1 - Line 313
            // Half Line - OK
            OC3R = SHORT_SYNC;
            OC3CON = 0x0009;    // single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = H_PAL - HSYNC_T;	//
            OC2CON = 0x0009;    		// single event 
			// on the last pulse prepare for the new field
            if ( VCount == 0)
            {   
               VPtr = VMap;
            }
            break;//	

  		case SV_SYNC_2://FIELD 2 - Lines 314,315
           // horizontal sync pulse - OK
           	OC3R = (H_PAL/2) - HSYNC_T;
            OC3CON = 0x0009;    // single event 

			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = H_PAL - HSYNC_T;		//
            OC2CON = 0x0009;    		// single event 
            break;

		case SV_POSTEQ_2://FIELD 2 - Lines 316,317
           	// Post equalization pulse - OK
            OC3R = SHORT_SYNC;
            OC3CON = 0x0009;    //single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = (H_PAL/2) + SHORT_SYNC;	// Relative to Timer 3
            OC2CON = 0x0009;    			// single event 
         	break;

		case SV_BLANKLINE_2://FIELD 2 - Lines vary
			// blank line - OK
            OC3R = HSYNC_T;
            OC3CON = 0x0009;    // single event 
            break;
            
		case SV_PREEQ_2://FIELD 2 - Lines 623-625
           // Pre equalization pulse - OK
            OC3R = SHORT_SYNC;
            OC3CON = 0x0009;    //single event 
			
			OC1R = (H_PAL/2);	//
            OC1CON = 0x0009;    // single event 

			OC2R = (H_PAL/2) + SHORT_SYNC;	// Relative to Timer 3
            OC2CON = 0x0009;    			// single event 

			// on the last pre-eq prepare for the new field
            if ( VCount == 0)
            {   
               VPtr = VMap;
            }
            break;

        default:            
        case SV_LINE_2://FIELD 2 - Lines vary
            // horizontal sync pulse - OK
            OC3R = HSYNC_T;
            OC3CON = 0x0009;    // single event 

            // activate OC4 for the SPI loading
            OC4R = HSYNC_T + BPORCH_T;
            OC4CON = 0x0009;    // single event 
            HCount = HRES/128;  // loads 8x16 bits at a time 
            break;                                        
    } //switch

    // advance the state machine
    if ( VCount == 0)
    {
		VCount = VC[ VState];
        VState = VS[ VState];
     }

    // clear the interrupt flag
    _T3IF = 0;
} // T3Interrupt

//-------------------------------------------START INTERRUPT ROUTINES----------------------------------------
void _ISRFAST _OC1Interrupt( void)//modified*
{
    SYNC = 0;   // bring the output down to SYNC level
    _OC1IF = 0; // clear the interrupt flag

} // OC1Interrupt

void _ISRFAST _OC2Interrupt( void)//modified*
{
    SYNC = 1;   // bring the output up to the black level
    _OC2IF = 0; // clear the interrupt flag

} // OC2Interrupt

void _ISRFAST _OC3Interrupt( void)
{
    SYNC = 1;   // bring the output up to the black level
    _OC3IF = 0; // clear the interrupt flag

} // OC3Interrupt

void _ISRFAST _OC4Interrupt( void) 
{    
    // load SPI FIFO with 8 x 16-bit words = 128 pixels
   	SPI1BUF = *VPtr++; 
   	SPI1BUF = *VPtr++; 
    SPI1BUF = *VPtr++; 
   	SPI1BUF = *VPtr++; 
   	SPI1BUF = *VPtr++; 
   	SPI1BUF = *VPtr++; 
   	SPI1BUF = *VPtr++; 
    SPI1BUF = *VPtr++; 
                            
    //if ( --HCount > 0)
   	//{   // activate again in time for the next SPI load
    //      OC4R += ( PIX_T * 7 * 16); 
    //     OC4CON = 0x0009;    // single event                 
	//}
	        
	// clear the interrupt flag
	_OC4IF = 0;

} // OC4Interrupt

//-------------------------------------------END INTERRUPT ROUTINES----------------------------------------

void initVideo( void)
{
    // set the priority levels
	_T3IP = 4; 		// this is the default value anyway
	_OC1IP = 4;		//modified*
    _OC2IP = 4;		//modified*
	_OC3IP = 4;
    _OC4IP = 4;

	TMR3 = 0;		// clear the timer
	PR3 = H_PAL;	// set the period register to PAL line 
	
	// 2.1 configure Timer3 modules
	T3CON = 0x8000;	// enabled, prescaler 1:1, internal clock 	
	
	// 2.2 init Timer3/OC1/OC2/OC3/OC4 Interrupts, clear the flag
	_OC1IF = 0;	_OC1IE = 1;		//modified*
    _OC2IF = 0; _OC2IE = 1;		//modified*
	_OC3IF = 0;	_OC3IE = 1;
    _OC4IF = 0; _OC4IE = 1;
    _T3IF = 0;  _T3IE = 1;
    	
	// 2.3 init the processor priority level
	_IPL = 0;	// this is the default value anyway

    // init the SPI1 
    if ( PIX_T == 2)
        SPI1CON1 = 0x043B;      // Master, 16 bit, disable SCK, disable SS, prescaler 1:2

    else if ( PIX_T == 3)
    	SPI1CON1 = 0x0437;      // Master, 16 bit, disable SCK, disable SS, prescaler 1:3

	else if ( PIX_T == 4)
    	SPI1CON1 = 0x0433;      // Master, 16 bit, disable SCK, disable SS, prescaler 1:4

	else if ( PIX_T == 5)
    	SPI1CON1 = 0x042f;      // Master, 16 bit, disable SCK, disable SS, prescaler 1:5

	else 
    	SPI1CON1 = 0x042b;      // Master, 16 bit, disable SCK, disable SS, prescaler 1:6

    SPI1CON2 = 0x0001;      // Enhanced mode, 8x FIFO
    SPI1STAT = 0x8000;      // enable SPI port
    
    // init PORTF for the Sync
    _TRISG0 = 0;            // output the SYNC pin
    	
	// init the vertical sync state machine
	VState = SV_PREEQ_2;
    VCount = PREEQ_N2;
    
} // initVideo


void clearScreen( void)
{
    int i;
    int *v;
    
    v = (int *)&VMap[0];
    
    // clear the screen     
    for ( i=0; i < (VRES*( HRES/16)); i++)
        *v++ = 0;
} //clearScreen


void haltVideo( void)
{
    T3CONbits.TON = 0;   // turn off the vertical state machine
} //haltVideo

void synchV( void) 
{
    while ( VCount != 1);
} // synchV


void plot( unsigned x, unsigned y) 
{
     if ((x<HRES) && (y<VRES) )
        VMap[ ((VRES-1-y)<<4) + (x>>4)] |= (0x8000 >> (x & 0xf));
} // plot

#define abs( a)     (((a)> 0) ? (a) : -(a))

void line( int x0, int y0, int x1, int y1)
{
     int steep, t ;
     int deltax, deltay, error;
     int x, y;
     int ystep;

     steep = ( abs(y1 - y0) > abs(x1 - x0));
     
     if ( steep )
     { // swap x and y
         t = x0; x0 = y0; y0 = t;
         t = x1; x1 = y1; y1 = t;
     }
     if (x0 > x1) 
     {  // swap ends
         t = x0; x0 = x1; x1 = t;
         t = y0; y0 = y1; y1 = t;
     }
     
     deltax = x1 - x0;
     deltay = abs(y1 - y0);
     error = 0;
     y = y0;
     
     if (y0 < y1) ystep = 1; else ystep = -1;
     for (x = x0; x < x1; x++)
     {
         if ( steep) plot(y,x); else plot(x,y);
         error += deltay;
         if ( (error<<1) >= deltax)
         {
             y += ystep;
             error -= deltax;
         } // if 
     } // for
} // line 